package cologne.eck.all_peas.control;

/*
 * Peafactory - Production of Password Encryption Archives
 * Copyright (C) 20221  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.awt.Component;
import java.io.File;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;

import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.all_peas.gui.PeaDialog;
import cologne.eck.tools.ExceptionHelper;
import cologne.eck.tools.StringNumberHelper;
import cologne.eck.tools.TestLog;
import cologne.eck.tools.UnexpectedValueException;
import cologne.eck.tools.WebTools;

public class UpdateChecker {
	
	private static File testFile = null;
	protected static void setTestFile(File _testFile) {	testFile = _testFile;}	
	
	/**
	 * Check if there is an update of the current PEA available
	 * (read from PAD file). 
	 * 
	 * @param c				the Component
	 * @param showErrors	true: show errors as dialogs
	 */
	public static void checkUpdates(Component c, boolean showErrors){
		
		TestLog.v(UpdateChecker.class, "Check updates... " );
		// create PAD file url: 
		// e.g. https://eck.cologne/peafactory/src/PAD/calendarPAD.xml
		String baseUrl = "https://eck.cologne/peafactory/src/PAD/";
		String peaExtension = null;
		String appType = PeaProperties.getFileType();
		if (appType.equals("calendar")) {
			peaExtension = "calendarPAD.xml";
		} else if (appType.equals("text file")) {
			peaExtension = "notebookPAD.xml";
		} else if (appType.equals("file")) {
			peaExtension = "filePAD.xml";
		} else if (appType.equals("image")) {
			peaExtension = "imagePAD.xml";
		} else {
			new UnexpectedValueException("String", "appType", "is not valid: " + appType).printDescription();
			if (showErrors == true) {
				PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("unexpected_error"), 
						PeaProperties.getVmBridge().translate("error"), 0);
			}
			return;
		}
		String urlString = baseUrl + peaExtension;
		if (WebTools.checkConnection() == false){
			if (showErrors == true) {
				PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("bad_web_connection"), 
						PeaProperties.getVmBridge().translate("error"), 0);
			}
			return;
		}
		TestLog.v(UpdateChecker.class, "Check URL: " + urlString);
		File file = WebTools.getFileFromURL(urlString);
		// for testing:
		if (testFile != null) file = testFile;
		
		if (file == null) {
			if (showErrors == true) {
				PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("download_failed"), 
						PeaProperties.getVmBridge().translate("error"), 0);
			}
			return;				
		} else {
			DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
					.newInstance();
			DocumentBuilder documentBuilder;
			try {
				documentBuilder = documentBuilderFactory.newDocumentBuilder();
				Document document = documentBuilder.parse(file);
				
				//TestLog.v(UpdateChecker.class, "Available version... " );
				String availableVersion = "";
				if (document != null
						&& document.getElementsByTagName("Program_Version") != null 
						&& document.getElementsByTagName("Program_Version").item(0) != null 
						&& document.getElementsByTagName("Program_Version").item(0).getTextContent() != null) {
					availableVersion = document.getElementsByTagName("Program_Version").item(0).getTextContent();
					if (availableVersion == null) {
						TestLog.e(UpdateChecker.class, "Missing available version...");
						if (showErrors == true) {
							PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("unexpected_error"), 
									PeaProperties.getVmBridge().translate("error"), 0);
						}
						return;
					}
				} else {
					TestLog.e(UpdateChecker.class, "Missing available version...");
					if (showErrors == true) {
						PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("unexpected_error"), 
								PeaProperties.getVmBridge().translate("error"), 0);						
					}
					return;
				}
				//TestLog.v(UpdateChecker.class, "Available version: " + availableVersion);
				// get release status: Major Update|Minor Update|New Release|Beta|Alpha|Media Only
				String releaseStatus = document.getElementsByTagName("Program_Release_Status").item(0).getTextContent();
				if (releaseStatus == null) {
					releaseStatus = "";
				}
				//TestLog.v(UpdateChecker.class, "Release: " + releaseStatus);
				String currentVersion = PeaControl.getDialog().getVersion();
				//String currentVersionNumber = currentVersion.replaceAll("[^0-9.]", "");
				TestLog.o(UpdateChecker.class, "current version: " + currentVersion
						+ ", available version: " + availableVersion + ", release status: " + releaseStatus);

				int newVersion = StringNumberHelper.checkNewVersionNumber(currentVersion, availableVersion);
				//TestLog.o(UpdateChecker.class, "newVersion: " + newVersion);
				if (newVersion == 1) {
					// always show new version
				} else if (newVersion == 0) {
					// in current version string contains possibly alpha or beta
					// check release status:
					if (currentVersion.toLowerCase().contains("alpha")) {
						if ( releaseStatus.toLowerCase().equals("alpha")) {
							if (showErrors == true) {
								PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("no_update_found"));
							}
							return;
						} else {
							// show new version
						}
					} else if (currentVersion.contains("beta")) {
						if ( releaseStatus.toLowerCase().equals("beta") || releaseStatus.toLowerCase().equals("alpha")) {
							if (showErrors == true) {
								PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("no_update_found"));
							}
							return;
						} else {
							// show new version
						}		
					} else {
						if (showErrors == true) {
							PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("no_update_found"));
						}
						return;
					}
				} else {
					if (showErrors == true) {
						PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("no_update_found"));
					}
					return;
				}
				// TODO date of new version, changes...
				String[] messageAndLink = {PeaProperties.getVmBridge().translate("web"), 
						"http://eck.cologne/peafactory/en/" + PeaControl.getDialog().getWebsite()
				};
				String[][] messagesAndLinks = {messageAndLink};
				PeaDialog.showMessageWithLink(c, 
						PeaProperties.getVmBridge().translate("update_found") + "\n      " 
								+ PeaProperties.getVmBridge().translate("version") + " " + availableVersion//av
									+ " (" + releaseStatus + ")",
								messagesAndLinks, null, -1);
			} catch (ParserConfigurationException e) {
				TestLog.e(UpdateChecker.class, new ExceptionHelper(e).getInfos());
				if (showErrors == true) {
					PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("unexpected_error"), 
							PeaProperties.getVmBridge().translate("error"), 0);
				}
			} catch (Exception e) {
				TestLog.e(UpdateChecker.class, e.toString() + " - " + e.getLocalizedMessage());

				if (showErrors == true) {
					PeaDialog.showMessage(c, PeaProperties.getVmBridge().translate("unexpected_error"), 
							PeaProperties.getVmBridge().translate("error"), 0);
				}
			}
		}
	}
}
