package cologne.eck.image_pea;

import java.nio.charset.Charset;
import java.util.Arrays;

import com.sun.imageio.plugins.png.PNGMetadata;

import cologne.eck.tools.Converter;
import cologne.eck.tools.ReadResources;
import cologne.eck.tools.TestLog;

class PNGMetadataHandler extends MetadataHandler {
	
	// Used StringBuilder for all methods
	private StringBuilder sb = null;
		
	// A PNG always starts with an 8-byte signature: 137 80 78 71 13 10 26 10 (decimal values)
	// hex: 89  50  4e  47  0d  0a  1a  0a
	private final static byte[] PNG_START = {(byte) 137, 80, 78, 71, 13, 10, 26, 10 };
	// chunks: identifiers for sections
	private final static byte[] END_CHUNK = "IEND".getBytes(Charset.forName("US-ASCII"));
	private final static byte[] tIME_CHUNK = "tIME".getBytes(Charset.forName("US-ASCII"));
	private final static byte[] tEXt_CHUNK = "tEXt".getBytes(Charset.forName("US-ASCII"));
	private final static byte[] zTXt_CHUNK = "zTXt".getBytes(Charset.forName("US-ASCII"));
	private final static byte[] iTXt_CHUNK = "iTXt".getBytes(Charset.forName("US-ASCII"));	
	private final static byte[] eXIf_CHUNK = "eXIf".getBytes(Charset.forName("US-ASCII"));
	
	// First part of an zTXt or iTxt chunk that contains raw data of a specific format
	private final static byte[] RAW_PROFILE = "Raw profile type ".getBytes(Charset.forName("US-ASCII"));
	// format names for raw profiles
	private final static byte[] exif_ASCII = "exif".getBytes(Charset.forName("US-ASCII")); // EXIF_Profile
	private final static byte[] EXIF_ASCII = "EXIF".getBytes(Charset.forName("US-ASCII")); // EXIF_Profile
	private final static byte[] iptc_ASCII = "iptc".getBytes(Charset.forName("US-ASCII")); // IPTC_Profile - IPTC IIM or Photoshop IRB format
	private final static byte[] _8bim_ASCII = "8bim".getBytes(Charset.forName("US-ASCII")); // Photoshop_Profile 
	private final static byte[] APP1_ASCII = "APP1".getBytes(Charset.forName("US-ASCII")); // APP1_Profile
	private final static byte[] icc_ASCII = "icc".getBytes(Charset.forName("US-ASCII")); // ICC_Profile
	private final static byte[] icm_ASCII = "icm".getBytes(Charset.forName("US-ASCII")); // ICC_Profile
	private final static byte[] xmp_ASCII = "xmp".getBytes(Charset.forName("US-ASCII")); // XMP_Profil
	

	
	// a series of chunks beginning with an IHDR chunk and ending with an IEND chunk
	
	// Each chunk consists of four parts: 
	// Length 	4 bytes 	unsigned integer giving the number of bytes in the chunk's data field
	// Chunk type 	4 bytes 	1. byte: chunk is critical = necessary for display (uppercase) or not (lowercase)
	//							2. byte: chunk is "public" (uppercase or private (lowercase)
	//							3. must be uppercase, reserved for future expansion.
	//							4. chunk is safe to copy by editors that do not recognize it
	// Chunk data 	Length bytes 
	// CRC 	4 bytes 	A CRC (Cyclic Redundancy Check) calculated on the preceding bytes in the chunk, chunk length is excluded
		
	// The iTXt (UTF-8), tEXt (ISO/IEC 8859-1 (Latin-1), uncompressed), and zTXt (compressed) chunks (text chunks) contain metadata of PNG file.
	// Each of the text chunks contains as its first field a keyword that indicates the type of information represented by the text string.

	// tEXt: Keyword: 1-79 bytes - 	Null separator: 1 byte - Text: n bytes 
	// zTXt: Keyword: 1-79 bytes -	Null separator: 1 byte - Compression method: 1 byte - Compressed text: n bytes 
	// iTXt: Keyword: 1-79 bytes -	Null separator: 1 byte - Compression flag: 1 byte 
	//		- Compression method: 1 byte - Language tag: 0 or more bytes - Null separator: 1 byte 
	//		- Translated keyword: 0 or more bytes  - Null separator: 1 byte - Text: n bytes 
	
	// The following keywords are predefined:
/*	Title 	Short (one line) title or caption for image
	Author 	Name of image's creator
	Description 	Description of image (possibly long)
	Copyright 	Copyright notice
	Creation Time 	Time of original image creation
	Software 	Software used to create the image
	Disclaimer 	Legal disclaimer
	Warning 	Warning of nature of content
	Source 	Device used to create the image
	Comment 	Miscellaneous comment; conversion from GIF comment */
	
	// an XMP packet is embedded in a PNG graphic file by adding a chunk of type iTXt with the keyword 'XML:com.adobe.xmp'.
	// when Exif, IPTC are added, they are stored in zTXt text chunks and save as ASCII.

	/* 4.2.4.6. tIME Image last-modification time
The tIME chunk gives the time of the last image modification (not the time of initial image creation). It contains:
   Year:   2 bytes (complete; for example, 1995, not 95)
   Month:  1 byte (1-12)
   Day:    1 byte (1-31)
   Hour:   1 byte (0-23)
   Minute: 1 byte (0-59)
   Second: 1 byte (0-60) 
Universal Time (UTC, also called GMT) should be specified rather than local time.  */
	
	/* FOR TESTING
	public static void main(String[] args) {		
//		PeaProperties.setWorkingMode("-t");
		String fileName = "/home/ax/TEST/Image/berberaffe-exif.png";
//		String fileName = "/home/ax/TEST/Image/img3.png";
		String result = new PNGMetadataHandler().getMetadata(fileName);		
		System.out.println(result);
	} */

	@Override
	public String getMetadata(byte[] imageBytes) {
		if (sb == null) {
			sb = new StringBuilder();
		}
		int len = imageBytes.length;
		byte[] start = new byte[8];
		System.arraycopy(imageBytes, 0, start,  0,  8);
		if (Arrays.equals(start,  PNG_START)) {
			TestLog.v(PNGMetadataHandler.class, "Is png");
		} else {
			sb.append("Is not a PNG file. \n");
			return new String(sb);			
		}
		// IHDR chunk:
		byte[] chunkLen = new byte[4];
		byte[] chunk = new byte[4];
		System.arraycopy(imageBytes, 8, chunkLen,  0,  4);
		System.arraycopy(imageBytes, 12, chunk,  0,  4);
		int lenInt = Converter.bytes2intsBE(chunkLen)[0];
		//System.out.println("	LEN: " + lenInt + " " + new String(chunk));
		if ( ! new String(chunk).equals("IHDR")) {
			TestLog.ve(PNGMetadataHandler.class, "Missing IHDR chunk as first chunk", 5);
		} 
		int index = 8 + 12 + 13; // 12 = 4 (len) + 4 (type)+ 4 (crc)
		String xmlString = null;
		while ( ! Arrays.equals(chunk, END_CHUNK)  && ! (index + 12 > len)) {
			System.arraycopy(imageBytes, index, chunkLen,  0,  4);
			System.arraycopy(imageBytes, index + 4, chunk,  0,  4);
			lenInt = Converter.bytes2intsBE(chunkLen)[0];
			// append values for tIME tEXt zTXt iTXt
			if (Arrays.equals(chunk, tIME_CHUNK)) {
				if (lenInt < 7) {
					TestLog.ve(PNGMetadataHandler.class, "Data too short to contain date and time: " + lenInt , 5);
					sb.append("Can't read last modified time. \n");
				} else {
					sb.append("\nLast modified: \n   Year: " + getBE(imageBytes[index + 8], imageBytes[index + 8 + 1]) 
						+ "\n   Month: " + imageBytes[index + 8 + 2] + "\n   Day: " + imageBytes[index + 8 + 3]
								+ "\n   Time: " + imageBytes[index + 8 + 4] + ":" + imageBytes[index + 8 + 5]);
				}

			} else if (Arrays.equals(chunk, tEXt_CHUNK)) {
				//System.out.println("tEXt");
				byte[] data = new byte[lenInt];
				System.arraycopy(imageBytes, index + 8, data,  0, lenInt);
				String dataString = new String(data, Charset.forName("ISO-8859-1"));
				sb.append("\n" + dataString + "\n");
				
			} else if (Arrays.equals(chunk, zTXt_CHUNK)) {
				//System.out.println("zTXt");
				if (proceedRawType(sb, index + 8, imageBytes, lenInt) == false) { // true: is processed in processRawType
				//Many pieces of software encode EXIF in a PNG by convention in an iTXt (or compressed zTXt) chunk notated "Raw profile type APP1".
				// There are no standard for Exif, IPTC data. In Exiv2, when Exif, IPTC are added, they are stored in zTXt text chunks and save as ASCII.

					byte[] data = new byte[lenInt];
					System.arraycopy(imageBytes, index + 8, data,  0, lenInt);
					String dataString = new String(data, Charset.forName("ISO-8859-1"));
					// there are sometimes many empty lines...
					int firstNewLine = dataString.indexOf("\n");
					if (dataString.indexOf("\n", firstNewLine) != -1) {
						String[] lines = dataString.split("\n");
						StringBuilder lineBuilder = new StringBuilder();
						for (String line : lines) {
							if (line.trim().length() > 0) {
								lineBuilder.append(line + "\n");
							}
						}
						dataString = new String(lineBuilder);
					}
					sb.append("\n" + dataString + "\n");
				}
			} else if (Arrays.equals(chunk, iTXt_CHUNK)) { // true: is processed in processRawType
				
				if (proceedRawType(sb, index,imageBytes, lenInt) == false) {
					byte[] data = new byte[lenInt];
					System.arraycopy(imageBytes, index + 8, data,  0, data.length);
					String dataString = new String(data, Charset.forName("UTF-8"));
					// there are sometimes many empty lines...
					int firstNewLine = dataString.indexOf("\n");
					if (dataString.indexOf("\n", firstNewLine) != -1) {
						String[] lines = dataString.split("\n");
						StringBuilder lineBuilder = new StringBuilder();
						for (String line : lines) {
							if (line.trim().length() > 0) {
								lineBuilder.append(line + "\n");
							}
						}
						dataString = new String(lineBuilder);
					}
					if (dataString.startsWith("XML")) { // append later
						xmlString = "\n\n" + dataString;
					} else {
						sb.append("\n" + dataString + "\n");// already new line from builder
					}
				}
			} else if (Arrays.equals(chunk, eXIf_CHUNK)) {
				byte[] data = new byte[lenInt];
				System.arraycopy(imageBytes, index + 8, data,  0, lenInt);
				TestLog.v(PNGMetadata.class, "eXIf chunk: ");
				//System.out.println("eXIf as hex:\n" +Converter.bytes2hex(data));
				String exifMetadata = new JPGMetadataHandler().getMetadataAllFormat(data, false);
				sb.append(exifMetadata);
			} 
			
			index += 12 + lenInt;
		}		
		// append XML string if present
		if (xmlString != null) {
			sb.append(xmlString);
		}
		return new String(sb);
	}

	/**
	 * Process an embedded raw type profile such as EXIF or IPTC
	 * 
	 * @param sb			the StringBuilder to append messages to
	 * @param index			index in image bytes
	 * @param imageBytes	image as byte array
	 * @param lenInt		length of data to proceed
	 * 
	 * @return	true if raw type was found, false otherwise
	 */
	private boolean proceedRawType(StringBuilder sb, int index, byte[] imageBytes, int lenInt) {
		if (sb == null) {
			sb = new StringBuilder();
		}
		// starts with fixed string
		byte [] rawTest = new byte[RAW_PROFILE.length];
		System.arraycopy(imageBytes, index, rawTest,  0, rawTest.length);
		if (Arrays.equals(rawTest, RAW_PROFILE)) {
			TestLog.v(PNGMetadataHandler.class, "Found raw profile type:");
			sb.append("Raw Profile: \n");
			byte[] rawType = new byte[4];
			System.arraycopy(imageBytes, index + RAW_PROFILE.length, rawType,  0, rawType.length);
			if (Arrays.equals(rawType, exif_ASCII) || Arrays.equals(rawType, EXIF_ASCII)) {
				TestLog.v(PNGMetadataHandler.class, "Found raw type: EXIF_Profile");
				sb.append("EXIF_Profile: Can't read\n");
				byte[] data = new byte[lenInt - RAW_PROFILE.length - rawType.length];
				System.arraycopy(imageBytes, index + RAW_PROFILE.length + rawType.length, data, 0, data.length);
				// TODO
				return true;
			} else if (Arrays.equals(rawType, iptc_ASCII)) {
				TestLog.v(PNGMetadataHandler.class, "Found raw type: IPTC_Profile - IPTC IIM or Photoshop IRB format");
				sb.append("IPTC_Profile: Can't read\n");
				byte[] data = new byte[lenInt - RAW_PROFILE.length - rawType.length];
				System.arraycopy(imageBytes, index + RAW_PROFILE.length + rawType.length, data, 0, data.length);
				// TODO
				return true;
			} else if (Arrays.equals(rawType, _8bim_ASCII)) {
				TestLog.v(PNGMetadataHandler.class, "Found raw type: Photoshop_Profile 8bim");
				sb.append("Photoshop_Profile 8bim: Can't read\n");
				return true;
			} else if (Arrays.equals(rawType, APP1_ASCII)) {
				TestLog.v(PNGMetadataHandler.class, "Found raw type: APP1_Profile");
				sb.append("APP1_Profile: Can't read\n");
				return true;
			} else {
				// 3-letter types: icc, icm, xmp
				rawType = new byte[3];
				System.arraycopy(imageBytes, index + RAW_PROFILE.length, rawType,  0, rawType.length);
				System.out.println("rawType: " + new String(rawType) );
				if (Arrays.equals(rawType, icc_ASCII) || Arrays.equals(rawType, icm_ASCII)) {
						TestLog.v(PNGMetadataHandler.class, "Found raw type: ICC_Profile");
						sb.append("ICC_Profile: Can't read\n");
						return true;
				} else if (Arrays.equals(rawType, xmp_ASCII)) {
					TestLog.v(PNGMetadataHandler.class, "Found raw type: ICC_Profile");
					sb.append("XMP_Profil: Can't read\n");
					return true;
				}					
			}			
		}
		return false;
	}
	/**
	 * Get short for big endian
	 * @param one	first (left) byte
	 * @param two	second byte
	 * @return	short for big endian
	 */
	private static int getBE(byte one, byte two) {
		return ((one & 0xff) << 8) | (two & 0xff);
	}
	
	@Override
	public String getMetadata(String fileName) {
		byte[] pngBytes = ReadResources.readExternFile(fileName);	
		return getMetadata(pngBytes);
	}

}
